# default lib imports
import logging
from geventwebsocket.websocket import WebSocket
from typing import List, Dict

# external lib imports
from flask import Flask
from flask_sockets import Sockets

# lib imports
from .IoTClient import IoTClient
from .DeviceType import DeviceType
from .error import ConnectionEnded
from .types import event_pair


# main server class
class IoTManager:
    """
    Main iot.io server for managing clients. Uses flask-sockets as the websocket endpoint. Requires a Flask instance
    to hook onto for the flask-sockets server.
    """

    def __init__(self, flask_app: Flask, logging_level: int = logging.ERROR, client_logging_level: int = logging.ERROR):
        # websocket server run off of a flask app
        self.sockets = Sockets(flask_app)

        # list of active clients
        self.__clients: List[IoTClient]
        self.__clients = []

        # a dict of device types
        self.__types: Dict[str, DeviceType]
        self.__types = {}

        # logger for the manager
        self.logger = logging.Logger("[iot.io.server]")

        # logging levels for main and client debuggers
        self.logger.setLevel(logging_level)
        self.client_logging_level = client_logging_level

        # route fo clients to connect to
        @self.sockets.route('/iot.io')
        def socket(ws: WebSocket):
            # create a new client
            try:
                # create a new client object if the client sends the handshake packet successfully
                client = IoTClient(ws, client_logging_level)
            except ConnectionEnded:
                # exit if the handshake fails
                return

            # check if the client has a valid type
            if self.__types.get(client.type, None) is None:
                # log and exit if the client has an invalid type
                self.logger.warning("Client with ID '" + client.id + "' claimed a device type of '" + client.type
                                    + "' but no such device type was defined. Refusing connection.")
                return

            # add the client to the client list
            self.__clients.append(client)

            # call the on_connect handler
            self.__on_connect_handlers(client)

            # begin the loop of accepting messages for the client
            client.loop(self.__handle_event_message)

            # call the on_disconnect handler
            self.__on_disconnect_handlers(client)

            # remove the client from the list of clients
            self.__clients = [x for x in self.__clients if client.id != x.__id]

    # handles the device specific and generic on_connect handlers for the specified client
    def __on_connect_handlers(self, client: IoTClient):
        """
        Execute the generic and device specific on_connect handler.

        :param client: New client object.
        :return:
        """

        try:
            # call the type specific handler
            self.__types[client.type].on_connect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling the type specific on_connect handler for client '" + client.id
                              + "': " + str(e))

        try:
            # call the generic handler
            self.on_connect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling generic on_connect handler for client '" + client.id
                              + "': " + str(e))

    # handles the device specific and generic on_disconnect handlers for the specified client
    def __on_disconnect_handlers(self, client: IoTClient):
        """
        Execute the generic and device specific on_disconnect handler.

        :param client: New client object.
        :return:
        """

        try:
            # call the type specific handler
            self.__types[client.type].on_disconnect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling the type specific on_disconnect handler for client '" + client.id
                              + "': " + str(e))

        try:
            # call the generic handler
            self.on_disconnect(client)
        except Exception as e:
            # log the exception
            self.logger.error("Error when calling generic on_disconnect handler for client '" + client.id
                              + "': " + str(e))

    # callback for when a client receives a message, sends it to the proper handler and squashes errors
    def __handle_event_message(self, event: str, message: str, client: IoTClient) -> event_pair:
        """
        Handles a given event and message given the client by executing the call_event_handler of the client's type.

        :param event: The event handler to trigger.
        :param message: The message to pass to the handler.
        :param client: The client object to pass to the handler.
        :return:
        """

        # activate the client type's event handler
        return self.__types[client.type].call_event_handler(event, message, client)

    # used to add a new device type to the manager
    def add_type(self, device: DeviceType):
        """
        Define a new device type for the Manager.

        :param device: Any implementation of the subclass DeviceType.
        :return: None
        """

        # ensure the device is of type DeviceType
        if not isinstance(device, DeviceType):
            raise ValueError("device is not of type DeviceType")

        # ensure device.type is of type str
        if not isinstance(device.type, str):
            raise ValueError("device.type is not of type str")

        # add the device to the list of device types
        self.__types.update({device.type: device})

        # logging output
        self.logger.debug("Successfully added DeviceType '" + device.type + "'.")

    # event decorator function
    def event(self, coroutine):
        """
        Decorator which allows for simple overwriting of the generic on_connect, and
        on_disconnect functions.

        :param coroutine: Function named on_connect, or on_disconnect.
        :return: bool
        """

        # handle general on_connect, and on_disconnect handlers
        if coroutine.__name__ == "on_connect" or coroutine.__name__ == "on_disconnect":
            # logging output
            self.logger.info("Event handler '" + coroutine.__name__ + "' was added successfully.")

            # replaces the existing coroutine with the provided one
            setattr(self, coroutine.__name__, coroutine)
            return True
        return False

    def on_connect(self, client: IoTClient):
        """
        A empty implementation of the generic on_connect function, meant to be overwritten.

        Runs when any device connects to the server.

        :param client: Client object representing the connecting client.
        :return: None or str.
        """
        pass

    def on_disconnect(self, client: IoTClient):
        """
        A empty implementation of the generic on_disconnect function, meant to be overwritten.

        Runs when any device disconnects from the server.

        :param client: Client object representing the disconnecting client.
        :return: None.
        """
        pass



# iot error base
class IoTError(Exception):
    """
    Base Exception for all iot.io errors
    """
    def __init__(self, message: str = "Base exception class for the iot.io library."):
        super().__init__(message)


# error when a client has its connection ended
class ConnectionEnded(IoTError):
    """
    Exception thrown when an operation attempts to use a socket but the connection is no longer alive.
    """
    def __init__(self, message: str = "Client socket no longer alive."):
        super().__init__(message)



# lib imports
from .IoTClient import IoTClient
from .types import event_pair


# superclass for device type definitions
class DeviceType:
    def __init__(self, type_name):
        self.type = type_name

    # calls the specific event handler if it exists
    def call_event_handler(self, event: str, message: str, client: IoTClient) -> event_pair:
        """
        Calls the event handler for the given event if it exists, if it doesnt exist then it does nothing.

        Don't overwrite.

        :param event: The event which is being invoked.
        :param message: The message to pass to the event handler.
        :param client: The client which sent the message.
        :return: None or str
        """
        try:
            # call the handler and check for a response
            response = getattr(self, "on_" + event)(message, client)

            print("response: " + response)

            if response is None:
                return None, None
            else:
                return event, response
        except AttributeError:
            return None, None

    def on_connect(self, client: IoTClient):
        pass

    def on_disconnect(self, client: IoTClient):
        pass



# default lib imports
from typing import Callable
import re
import json
import logging

# external lib imports
from geventwebsocket.websocket import WebSocket
from gevent import Timeout

# lib imports
from .error import ConnectionEnded
from .types import event_pair, sendable


# iot client class
class IoTClient:
    def __init__(self, ws: WebSocket, logging_level: int = logging.ERROR):
        # logger for the new client
        self.logger = logging.Logger("[iot.io.client:New Client]")
        self.logger.level = logging_level

        # reference to the client's websocket connection
        self.socket = ws

        # receive, waiting for the client to send the handshake packet, 30 second timeout before the connection closes
        with Timeout(30, False):
            # get the message
            message = ws.receive()

            # parse it
            self.__id, self.__type, self.__data = self.handshake(message)

            # set the new logger name
            self.logger.name = "[iot.io.client:" + self.__id + "]"

    # properties for id, type, and data
    @property
    def id(self):
        return self.__id

    @property
    def type(self):
        return self.__type

    @property
    def data(self):
        return self.__data

    # parse a iot.io packet and return the contents as a list
    @staticmethod
    def parse(message: str) -> list:
        """
        Messages between the client and server are formatted as a list, which each
         element being an argument. For example the handshake packet the client first sends to the server
         is structured like the following list: (id, type, data).

         This same packet would be sent as the following string by the client:
            '[/"1"/][/"id"/][/"type"/][/"data"/]'

        This parser takes than string and converts it to the following list:
            ['1', 'id', 'type', 'data']

        :param message: the message to be parsed as a string
        :return: a list of parsed arguments
        """

        # parse the message to return contents
        result = re.findall("\[/\"([^[\]]*)\"/\]", message)

        return result

    # parse a handshake packet, handle errors and notify the client
    def handshake(self, message: str):
        # parse the packet first
        args = self.parse(message)

        if len(args) != 3:
            # send an error message to client and end the connection
            self.logger.warning("Invalid handshake message, ending connection.")
            # TODO figure out an error system to notify clients
            self.socket.close()

            # raise the connection end error
            raise ConnectionEnded

        return args[0], args[1], args[2]

    # parse a normal message packet
    def parse_message(self, packet: str) -> event_pair:
        """
        Parse what is expected to be a basic iot.io message, returning the event and message

        :param packet: The iot.io packet to parse.
        :return: A event pair. [(None, None) or (str, sendable)]
        """
        # parse the packet
        args = self.parse(packet)

        # check for correct length
        if len(args) != 2:
            self.logger.warning("When attempting to parse a client message a invalid message format was received. ("
                                "Raw Message: " + str(packet) + " | Parsed Message: " + str(args) + ")")
            return None, None

        # attempt deserialization of message
        try:
            args[1] = json.loads(args[1])
        except ValueError or json.JSONDecodeError:
            pass

        # return the event and the message
        return args[0], args[1]

    # sends event message to the client
    def send(self, event: str, message: sendable):
        # check type of message
        if not isinstance(message, (bool, int, float, str, dict)):
            raise ValueError("A non-sendable value was passed to a client's send method.")

        # conversions from sendable to string
        if isinstance(message, dict):
            # serialize the dict as a json string
            message = json.dumps(message)
        # convert generics to string counterpart
        else:
            # convert to string
            message = str(message)

        # check if the connection is alive
        if not self.socket.closed:
            # logging
            self.logger.debug("Sending a message for event '" + event + "':" + message)

            # send the message
            self.socket.send("[/\"" + event + "\"/][/\"" + message + "\"/]")
        else:
            raise ConnectionEnded

    # loops and receives messages from the client and calls the specified callback when one is received
    def loop(self, callback: Callable[[str, sendable, 'IoTClient'], event_pair]):
        """
        Runs a loop which accepts client messages until the client closes the connection.

        :param callback: A callback function which accepts the event, message, and Client.
        :return: None
        """

        # loop while the connection is still alive
        while not self.socket.closed:
            # receive from the client
            message = self.socket.receive()

            # if the message is None then client connection has ended so exit the loop
            if message is None:
                break

            # parse the message
            event, message = self.parse_message(message)

            # check if even is none
            if event is not None:
                # call the callback
                event, response = callback(event, message, self)

                # check if there is a response to send to the client
                if response is not None:
                    try:
                        self.send(event, response)
                    except ConnectionEnded:
                        break
